//-----------------------------------------------------------------------------
// File:	RacorX.cpp
//
// Desc: Simple quad transformed by a vertex shader program. It uses the Common 
//       File routines
//
//		Last modification: January 5th, 2002
//
// Copyright (c) 1995-2000 Microsoft Corporation. All rights reserved.
// Copyright (c) 2001 - 2002 Wolfgang Engel wolf@direct3d.net All rights reserved.
//-----------------------------------------------------------------------------
#define STRICT
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFont.h"
#include "DXUtil.h"

#include "resource.h"


//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
// A structure for vertex type
struct VERTEX
{
    FLOAT x, y, z;      // The untransformed position for the vertex
};


//---------------------------------------------------------------------------
//shader decl
//---------------------------------------------------------------------------
DWORD dwDecl[] =
{
    D3DVSD_STREAM(0),
    D3DVSD_REG(0, D3DVSDT_FLOAT3 ),			//D3DVSDE_POSITION,0  
    D3DVSD_END()
};

//----------------------------------------------------------------------------
//      reg c4-7	= WorldViewProj matrix
//      reg c8      = constant color
//	    reg v0		= input register
//----------------------------------------------------------------------------
const char BasicVertexShader[] =
"vs.1.1             //Shader version 1.1			\n"\
"dp4 oPos.x, v0, c4 //emit projected position	    \n"\
"dp4 oPos.y, v0, c5 //emit projected position	    \n"\
"dp4 oPos.z, v0, c6 //emit projected position	    \n"\
"dp4 oPos.w, v0, c7 //emit projected position	    \n"\
"mov oD0, c8		//Diffuse color = c8            \n";


//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
protected:
    CD3DFont*     m_pFont;				// Font for drawing tex

    D3DXMATRIX	m_matProj;
    D3DXMATRIX	m_matView;
    D3DXMATRIX	m_matWorld;

    DWORD					m_dwVertexShader;	// vertex shader handle
	
	LPDIRECT3DVERTEXBUFFER8 m_pVB;				// vertex buffer
	DWORD					m_dwSizeofVertices;	// size of vertex buffer
	LPDIRECT3DINDEXBUFFER8  m_pIB;				// index buffer
	DWORD					m_wSizeofIndices;	// size of index buffer

    HRESULT ConfirmDevice( D3DCAPS8*, DWORD, D3DFORMAT );

    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT FinalCleanup();
    HRESULT Render();
    HRESULT FrameMove();

public:
    CMyD3DApplication();
};

CMyD3DApplication g_d3dApp;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{

    if( FAILED( g_d3dApp.Create( hInst ) ) )
        return 0;

    return g_d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
    m_strWindowTitle    = _T("RacorX");
    m_dwCreationWidth		= 800;   // Width used to create window
    m_dwCreationHeight		= 600;
    m_bUseDepthBuffer   = TRUE;
    m_pFont        = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
    m_pVB  = NULL;
	m_pIB  = NULL;
	m_dwVertexShader	= 0xffffffff;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    // rotates the object about the y-axis
    D3DXMatrixRotationY( &m_matWorld, m_fTime * 1.5f);

	// set the clip matrix
    D3DXMATRIX matTemp;
	D3DXMatrixTranspose( &matTemp , &(m_matWorld * m_matView * m_matProj) );
	m_pd3dDevice->SetVertexShaderConstant(4, matTemp, 4);
	
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
    // Clear the viewport | z-buffer
    m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
		D3DCOLOR_XRGB(0,0,128), 1.0f, 0L );

    // Begin the scene
    if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
    {
		// set the vertex buffer
		// == specify the source of stream 0
		m_pd3dDevice->SetStreamSource( 0, m_pVB, sizeof(VERTEX));

		// set the vertex shader
		m_pd3dDevice->SetVertexShader( m_dwVertexShader );	

		// set the index buffer
		m_pd3dDevice->SetIndices(m_pIB, 0);
	
		// ... rendering
		m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
											0,
											4,		// number of vertices
											0,
											2);		// number of primitives

        // Output statistics
        m_pFont->DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
        m_pFont->DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats );

        // End the scene.
        m_pd3dDevice->EndScene();
    }
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Initialize the font's internal textures
    m_pFont->InitDeviceObjects( m_pd3dDevice );

	// create and fill the vertex buffer
	// Initialize vertices for rendering a quad
	VERTEX Vertices[] =
	{
		//  x     y     z   
		{ -1.0f,-1.0f, 0.0f, },
		{  1.0f,-1.0f, 0.0f, },
		{  1.0f, 1.0f, 0.0f, },
		{ -1.0f, 1.0f, 0.0f, },
	};

	m_dwSizeofVertices = sizeof (Vertices);

	// Declare custom FVF macro.
	#define D3DFVF_VERTEX (D3DFVF_XYZ)
	
	// Create the vertex buffers with four vertices
    if( FAILED( m_pd3dDevice->CreateVertexBuffer( 4 * sizeof(VERTEX),
		D3DUSAGE_WRITEONLY, D3DFVF_VERTEX, D3DPOOL_MANAGED, &m_pVB ) ) )
       return E_FAIL;
	
    // lock and unlock the vertex buffer to fill it with memcpy
    VOID* pVertices;
    if( FAILED( m_pVB->Lock( 0, m_dwSizeofVertices, (BYTE**)&pVertices, 0 ) ) )
        return E_FAIL;
    memcpy( pVertices, Vertices, m_dwSizeofVertices);
    m_pVB->Unlock();

	// create and fill the index buffer
	// indices
	WORD wIndices[]={0, 1, 2, 0, 2, 3};

	m_wSizeofIndices = sizeof (wIndices);

	// create index buffer
	if(FAILED (m_pd3dDevice->CreateIndexBuffer(m_wSizeofIndices, 0, 
								D3DFMT_INDEX16, D3DPOOL_MANAGED, &m_pIB)))
		return E_FAIL;

	// fill index buffer
	VOID *pIndices;
	if (FAILED(m_pIB->Lock(0, m_wSizeofIndices, (BYTE **)&pIndices, 0)))
		return E_FAIL;
	memcpy(pIndices, wIndices, m_wSizeofIndices);
	m_pIB->Unlock();

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // Restore the device objects for the font
    m_pFont->RestoreDeviceObjects();

	// Set the projection matrix
    FLOAT fAspect = m_d3dsdBackBuffer.Width / (FLOAT)m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &m_matProj, D3DX_PI/4, fAspect, 1.0f, 100.0f );
	
	// Set up the view matrix. 
    D3DXMatrixLookAtLH( &m_matView, &D3DXVECTOR3( 0.0f, 1.0f,-4.0f ), //from
		&D3DXVECTOR3( 0.0f, 0.0f, 0.0f ), //at
		&D3DXVECTOR3( 0.0f, 1.0f, 0.0f ));//up

	// assemble and create the vertex shader
	char szBuffer[128];
	ID3DXBuffer*	pVS;
	ID3DXBuffer*	pErrors;

	HRESULT rc;
	
	// Assemble the shader	
	rc = D3DXAssembleShader( BasicVertexShader , sizeof(BasicVertexShader) -1, 0, NULL, &pVS ,
							 &pErrors );

	if ( FAILED(rc) )
	{
		OutputDebugString( "Failed to assemble the vertex shader, errors:\n" );
		OutputDebugString( (char*)pErrors->GetBufferPointer() );
		OutputDebugString( "\n" );
	}

	// Create the vertex shader
    rc = m_pd3dDevice->CreateVertexShader( dwDecl, (DWORD*)pVS->GetBufferPointer(),
		&m_dwVertexShader, 0 );
	if ( FAILED(rc) )
	{
		OutputDebugString( "Failed to create the vertex shader, errors:\n" );
		D3DXGetErrorString(rc,szBuffer,sizeof(szBuffer));
		OutputDebugString( szBuffer );
		OutputDebugString( "\n" );
	}
	SAFE_RELEASE(pVS);
	if ( FAILED(rc) )
		return rc;

	//set material color
	FLOAT fMaterial[4] = {0,1,0,0};
	m_pd3dDevice->SetVertexShaderConstant(8, fMaterial, 1);

	// z-buffer enabled
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );
	
    // Turn off D3D lighting, since we are providing our own vertex shader lighting
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );
	
	// Turn off culling, so we see the front and back of the quad
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
	
   return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{    
    m_pFont->InvalidateDeviceObjects();

	if ( m_dwVertexShader != 0xffffffff )
	{
		m_pd3dDevice->DeleteVertexShader( m_dwVertexShader );
		m_dwVertexShader = 0xffffffff;
	}
	
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{    
	m_pFont->DeleteDeviceObjects();

    SAFE_RELEASE( m_pVB );
    SAFE_RELEASE( m_pIB );

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
	SAFE_DELETE( m_pFont );

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{    
	if( (dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING ) ||
        (dwBehavior & D3DCREATE_MIXED_VERTEXPROCESSING ) )
    {
        if( pCaps->VertexShaderVersion < D3DVS_VERSION(1,1) )
            return E_FAIL;
    }

    return S_OK;
}

