//-----------------------------------------------------------------------------
// File:	RacorX.cpp
//
// Desc:	Bezier patch field transformed and lit by a vertex shader, that is 
//			compiled with NVASM. 
//			Shows diffuse/specular lighting
//
//			This example uses the Common File routines.
//
//			Last modification: February 9th, 2002
//
// Copyright (c) 1995-2000 Microsoft Corporation. All rights reserved.
// Copyright (c) 2002 wolf@direct3d.net All rights reserved.
//-----------------------------------------------------------------------------
#define STRICT
#include "stdio.h"
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFont.h"
#include "DXUtil.h"
#include "D3DUtil.h"

#include "resource.h"
#include "Const.h"			// constant register memory layout
#include "BPatch.h"

#define WM_MOUSEWHEEL                   0x020A



//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
protected:
    CD3DFont*   m_pFont;				// Font for drawing tex
    CD3DBPatch* m_pBPatch;	

    D3DXMATRIX	m_matProj, m_matView,
				m_matWorld,	m_matWorldI;

    D3DXVECTOR3 m_vObjectCenter;      // Center of bounding sphere of object
    FLOAT       m_fObjectRadius;      // Radius of bounding sphere of object
	
    CD3DArcBall m_ArcBall;            // trackball rotation utility
    FLOAT		m_fZDist; 

    D3DXVECTOR4 m_vLightColor[3];
    DWORD m_dwCurrentColor;	

    DWORD		m_dwVertexShader;	// vertex shader handle

    LPDIRECT3DTEXTURE8 m_pBackgroundTexture;

    BYTE m_bKey[256];			// keyboard state buffer

    HRESULT ConfirmDevice( D3DCAPS8*, DWORD, D3DFORMAT );

    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT FinalCleanup();
    HRESULT Render();
    HRESULT FrameMove();
	LRESULT MsgProc( HWND hWnd , UINT msg , WPARAM wParam , LPARAM lParam ); 

	HRESULT CreateVSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
								DWORD* dwDeclaration,
								TCHAR* strVSPath, 
								DWORD* m_dwVertexShader);


public:
    CMyD3DApplication();
};

CMyD3DApplication g_d3dApp;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    if( FAILED( g_d3dApp.Create( hInst ) ) )
        return 0;

    return g_d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
    m_strWindowTitle    = _T("RacorX #4");
	m_dwCreationWidth			= 800;				// Width used to create window
	m_dwCreationHeight			= 600;				// Height used to create window
    m_bUseDepthBuffer   = TRUE;
    m_pFont        = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
	m_pBackgroundTexture = NULL;
	m_dwVertexShader	= 0xffffffff;
    m_pBPatch = new CD3DBPatch();	
	m_fObjectRadius = 1;
	ZeroMemory( m_bKey, 256 );	

	// specular light colors
    m_vLightColor[0]=D3DXVECTOR4(0.3f,0.1f,0.1f,1.0f);
    m_vLightColor[1]=D3DXVECTOR4(0.1f,0.5f,0.1f,1.0f);
    m_vLightColor[2]=D3DXVECTOR4(0.0f,0.1f,0.4f,1.0f);
    m_dwCurrentColor=0;
	
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // Set cursor to indicate that user can move the object with the mouse
	#ifdef _WIN64
		SetClassLongPtr( m_hWnd, GCLP_HCURSOR, (LONG_PTR)LoadCursor( NULL, IDC_SIZEALL ) );
	#else
		SetClassLong( m_hWnd, GCL_HCURSOR, (LONG)LoadCursor( NULL, IDC_SIZEALL ) );
	#endif
	
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    // rotates and translate the object
    D3DXMatrixTranslation( &m_matWorld, -m_vObjectCenter.x,	-m_vObjectCenter.y,
							-m_vObjectCenter.z);
	
    D3DXMatrixMultiply( &m_matWorld, &m_matWorld, m_ArcBall.GetTranslationMatrix() );
    D3DXMatrixMultiply( &m_matWorld, &m_matWorld, m_ArcBall.GetRotationMatrix() );

	//Set up the view matrix. 
	D3DXVECTOR3 vEyePt (0.0f, 0.0f, -3.7f * m_fObjectRadius  + m_fZDist );
	D3DXVECTOR3 vLookAtPt (0.0f, 0.0f, 0.0f);
	D3DXVECTOR3 vUp (0.0f, 1.0f, 0.0f);
    D3DXMatrixLookAtLH(&m_matView, &vEyePt, &vLookAtPt, &vUp);	
	
	// set the clip matrix
	m_pd3dDevice->SetVertexShaderConstant(CLIP_MATRIX, &(m_matWorld * m_matView * m_matProj),4);

	// set the world inverse matrix
    D3DXMATRIX matWorldInverse;
    D3DXMatrixInverse(&matWorldInverse, NULL, &m_matWorld);
    m_pd3dDevice->SetVertexShaderConstant(INVERSE_WORLD_MATRIX, matWorldInverse,3);

	// stuff for specular lighting
	// set eye vector E
    m_pd3dDevice->SetVertexShaderConstant(EYE_VECTOR, vEyePt,1);	

	// specular color
    if(m_bKey['C']) 
	{
        m_bKey['C']=0;
        ++m_dwCurrentColor;
        if(m_dwCurrentColor >= 3) 
			m_dwCurrentColor=0;
    }	
	m_pd3dDevice->SetVertexShaderConstant(SPEC_COLOR, m_vLightColor[m_dwCurrentColor],1);
   return S_OK;
}


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
    // Clear the viewport | z-buffer
    m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 
		D3DCOLOR_XRGB(0,0,128), 1.0f, 0L );

    // Begin the scene
    if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
    {
		// set the vertex shader
		m_pd3dDevice->SetVertexShader( m_dwVertexShader );
		
		// set the texture
		m_pd3dDevice->SetTexture( 0, m_pBackgroundTexture );

		m_pBPatch->Render(m_pd3dDevice);

        // Output statistics
        m_pFont->DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );
        m_pFont->DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats );

        // End the scene.
        m_pd3dDevice->EndScene();
    }
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Initialize the font's internal textures
    m_pFont->InitDeviceObjects( m_pd3dDevice );

 	// shader decl
	DWORD dwDecl[] =
	{
		D3DVSD_STREAM(0),
		D3DVSD_REG(0, D3DVSDT_FLOAT3 ),	// input register v0  
		D3DVSD_REG(3, D3DVSDT_FLOAT3 ),	// normal in input register v3
		D3DVSD_REG(7, D3DVSDT_FLOAT2),	// tex coordinates
		D3DVSD_END()
	};

	// loads a *.vso binary file, already compiled with NVASM and creates a vertex shader
	if (FAILED(CreateVSFromCompiledFile (m_pd3dDevice, dwDecl,"shaders/specular.vso", &m_dwVertexShader)))
		return E_FAIL;
	
    // Load the texture for the background image
    if( FAILED( D3DUtil_CreateTexture( m_pd3dDevice, _T("ShaderX.tga"),
		&m_pBackgroundTexture) ) )
        return D3DAPPERR_MEDIANOTFOUND;
	
    VERTICES cp[9];
    cp[0].vPosition=D3DXVECTOR3(-1,1,0);	 cp[0].uv=D3DXVECTOR2(0.0f,0.0f);
    cp[1].vPosition=D3DXVECTOR3(0,1,0);      cp[1].uv=D3DXVECTOR2(0.5f,0.0f);
    cp[2].vPosition=D3DXVECTOR3(1,1,0);      cp[2].uv=D3DXVECTOR2(1.0f,0.0f);
	
    cp[3].vPosition=D3DXVECTOR3(-1,0,0);     cp[3].uv=D3DXVECTOR2(0.0f,0.5f);
    cp[4].vPosition=D3DXVECTOR3(0,0,-2);     cp[4].uv=D3DXVECTOR2(0.5f,0.5f);
    cp[5].vPosition=D3DXVECTOR3(1,0,0);      cp[5].uv=D3DXVECTOR2(1.0f,0.5f);
	
    cp[6].vPosition=D3DXVECTOR3(-1,-1,0);    cp[6].uv=D3DXVECTOR2(0.0f,1.0f);
    cp[7].vPosition=D3DXVECTOR3(0,-1,0);     cp[7].uv=D3DXVECTOR2(0.5f,1.0f);
    cp[8].vPosition=D3DXVECTOR3(1,-1,0);     cp[8].uv=D3DXVECTOR2(1.0f,1.0f);

	m_pBPatch->m_dwULevel = 4;
	m_pBPatch->m_dwVLevel = 4;
	m_pBPatch->Create(cp, 3, 3);
		
	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // Restore the device objects for the font
    m_pFont->RestoreDeviceObjects();	
    m_pBPatch->RestoreDeviceObjects(m_pd3dDevice);		

    // Set the projection matrix
    FLOAT fAspect = m_d3dsdBackBuffer.Width / (FLOAT)m_d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH( &m_matProj, D3DX_PI/4, fAspect, m_fObjectRadius/64.0f,
		m_fObjectRadius*200.0f);

	// set arc ball
    m_ArcBall.SetWindow( m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85f );
    m_ArcBall.SetRadius( m_fObjectRadius );
	
	// dithering on
    m_pd3dDevice->SetRenderState( D3DRS_DITHERENABLE, TRUE );

	// z-buffer enabled
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

    // Turn off D3D lighting, since we are providing our own vertex shader lighting
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );

    // Turn off culling, so we see the front and back of the quad
    m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

	// enable specular lighting
    m_pd3dDevice->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

	// texture settings
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP );
	
    // specular power
    m_pd3dDevice->SetVertexShaderConstant(SPEC_POWER,D3DXVECTOR4(0,10,25,50),1);

	// light direction
	D3DXVECTOR3 vLight(0,0,1);
	m_pd3dDevice->SetVertexShaderConstant(LIGHT_VECTOR, vLight,1);

	D3DXCOLOR matDiffuse(0.9f, 0.9f, 0.9f, 1.0f);
	m_pd3dDevice->SetVertexShaderConstant(DIFFUSE_COLOR, &matDiffuse, 1);

	
    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{    
    m_pFont->InvalidateDeviceObjects();
    m_pBPatch->InvalidateDeviceObjects();			

    return S_OK;
}


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{    
	m_pFont->DeleteDeviceObjects();
    m_pBPatch->DeleteDeviceObjects();			
	
    SAFE_RELEASE( m_pBackgroundTexture );
	
	if ( m_dwVertexShader != 0xffffffff )
	{
		m_pd3dDevice->DeleteVertexShader( m_dwVertexShader );
		m_dwVertexShader = 0xffffffff;
	}
	
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
	SAFE_DELETE( m_pFont );
	SAFE_DELETE( m_pBPatch );

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{    
	if( (dwBehavior & D3DCREATE_HARDWARE_VERTEXPROCESSING ) ||
        (dwBehavior & D3DCREATE_MIXED_VERTEXPROCESSING ) )
    {
        if( pCaps->VertexShaderVersion < D3DVS_VERSION(1,1) )
            return E_FAIL;
    }

    return S_OK;
}



//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: 
//
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    // Pass mouse messages to the ArcBall so it can build internal matrices
    m_ArcBall.HandleMouseMessages( hWnd, msg, wParam, lParam );

	switch (msg)
	{
	case WM_KEYDOWN:
        m_bKey[wParam] = 1;
		break;

	case WM_KEYUP:
        m_bKey[wParam] = 0;
		break;
	
	case WM_MOUSEWHEEL:
		short sr = HIWORD(wParam);
		m_fZDist -= (sr/120) * m_fObjectRadius;
		break;
	
	}
	
 return CD3DApplication::MsgProc(hWnd, msg, wParam, lParam);
}

//------------------------------------------------------------------------------
// Name:	CreateVSFromBinFile
// Desc:	loads a binary *.vso file that was compiled by NVASM
//			and creates a vertex shader
//------------------------------------------------------------------------------
HRESULT CMyD3DApplication::CreateVSFromCompiledFile (IDirect3DDevice8* m_pd3dDevice, 
													 DWORD* dwDeclaration,
													 TCHAR* strVSPath, 
													 DWORD* m_dwVS)
{
	char szBuffer[128];		// debug output
	DWORD*	dwpVS;			// pointer to address space of the calling process
	HANDLE hFile, hMap;		// handle file and handle mapped file
    TCHAR tempVSPath[512];	// temporary file path
	HRESULT hr;				// error 

    if( FAILED( hr = DXUtil_FindMediaFile( tempVSPath, strVSPath ) ) )
        return D3DAPPERR_MEDIANOTFOUND;
	
	hFile = CreateFile(tempVSPath, GENERIC_READ,0,0,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,0);

	if(hFile != INVALID_HANDLE_VALUE) 
	{
		if(GetFileSize(hFile,0) > 0) 
			hMap = CreateFileMapping(hFile,0,PAGE_READONLY,0,0,0);
		else
		{
			CloseHandle(hFile);
			return E_FAIL;		
		}
	}	
	else
		return E_FAIL;	
	
	// maps a view of a file into the address space of the calling process
	dwpVS = (DWORD *)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
		
	// Create the vertex shader
	hr = m_pd3dDevice->CreateVertexShader( dwDeclaration, dwpVS, m_dwVS, 0 );
	if ( FAILED(hr) )
	{
		OutputDebugString( "Failed to create Vertex Shader, errors:\n" );
		D3DXGetErrorStringA(hr,szBuffer,sizeof(szBuffer));
		OutputDebugString( szBuffer );
		OutputDebugString( "\n" );
	  return hr;
	}
	
	UnmapViewOfFile(dwpVS);
	CloseHandle(hMap);
	CloseHandle(hFile);
	
  return S_OK;
}